import numpy as np
import pandas as pd
from numpy import fft
from numba import jit
from scipy import signal


@jit(parallel=True)
def computeFourrierTransformCol(x, N):
    # x_freqdom = fft.fft(x)
    # return abs(x_freqdom), np.angle(x_freqdom)
    freq, dsp = signal.periodogram(x, nfft=N)
    return freq[-1], np.average(dsp)


@jit(parallel=True)
def computeFourrierTransformLine(x):
    # t = np.arange(0, N)
    # p = np.polyfit(t, x, 1)  # find linear trend in x
    # x_notrend = x - p[0] * t  # detrended x
    # x_freqdom = fft.fft(x_notrend)
    x_freqdom = fft.fft(x)
    return abs(x_freqdom), np.angle(x_freqdom)


class FourierFeaturesCol:
    def __init__(self, fft_period=21):
        self.fft_period = fft_period

    @jit
    def fourierDecomposition(self, X):
        T, N = X.shape
        predictors = np.array([])
        for me_col in range(N):
            x = X[:, me_col]
            max_freq, avg_dsp = computeFourrierTransformCol(x, T)
            predictors = np.append(predictors, np.hstack([max_freq, avg_dsp]))
        assert len(predictors) == N * 2
        return predictors

    @jit
    def featuresCreation(self, X, display_threshold=0.99):
        X_nd = X.values
        T, N = X_nd.shape
        new_features_length = N * 2
        X_final = np.empty((T, new_features_length))
        X_final[:] = np.nan
        for t in range(self.fft_period, T):
            if np.random.rand() >= display_threshold:
                print(str(t) + ' over ' + str(T))
            X_final[t, :] = self.fourierDecomposition(X_nd[t - self.fft_period:t, :])
        fft_features = pd.DataFrame(data=X_final, index=X.index,
                                    columns=['fff_' + str(i) for i in range(new_features_length)])

        return fft_features


class FourierFeaturesLine:
    def __init__(self):
        self.fft_features = pd.DataFrame()

    @jit
    def featuresCreation(self, X, display_threshold=0.99):
        col_names = list(X)
        amp_col_names = ['amp_lin_' + col for col in col_names]
        phase_col_names = ['phase_lin_' + col for col in col_names]
        new_col_names = amp_col_names + phase_col_names
        X_nd = X.values
        # T, N = X_nd.shape
        T = len(X_nd)
        N = len(X_nd[0])
        X_fft_col = np.zeros((T, 2 * N))
        for me_row in range(T):
            if np.random.rand() >= display_threshold:
                print(str(me_row) + ' over ' + str(T))
            x = X_nd[me_row, :]
            # t = np.arange(0, N)
            # p = np.polyfit(t, x, 1)  # find linear trend in x
            # x_notrend = x - p[0] * t  # detrended x
            # x_freqdom = fft.fft(x_notrend)
            # X_fft_col[me_row,:] = np.concatenate((abs(x_freqdom), np.angle(x_freqdom)))
            X_fft_col[me_row, :] = np.concatenate(computeFourrierTransformLine(x))
        return pd.DataFrame(data=X_fft_col, index=X.index, columns=new_col_names)


#!/usr/bin/env python3
# coding: utf-8

# Built-in packages

# Third party packages
import pandas as pd

# Local packages


__all__ = ['roll_corr']


def roll_corr(df, window=None, min_periods=2, **kwargs):
    """ Compute the rolling cross correlation.

    Parameters
    ----------
    df : pd.DataFrame
        Data to compute rolling correlation. Each columns must be a variable
        and each row is an observation.
    window : int, or offset
        Size of the moving window.
        - If its an int then this will be the number of observations used for
            calculating the statistic. Each window will be a fixed size.
        - If its an offset then this will be the time period of each window.
            Each window will be a variable sized based on the observations
            included in the time-period. This is only valid for datetimelike
            indexes.
        - Default is None, each window take all the past observations.
    min_periods : int, optional
        Minimum number of observations in window required to have a value
        (otherwise result is np.nan). Default is 2.
    **kwargs
        See pd.rolling for optional parameters.

    Returns
    -------
    pd.DataFrame
        Each column correspond to the rolling pairwise correlation.

    Warnings
    --------
    Not optimized function, may be time expensive for large dataset.

    """
    if window is None:
        window = df.index.size

    col, col2 = df.columns, list(df.columns)
    mat_roll_cor = df.rolling(window, min_periods=min_periods, **kwargs).corr()
    flat_roll_cor = pd.DataFrame(index=df.index)

    for c in col:
        col2.remove(c)
        temp = mat_roll_cor.loc[:, c].unstack()

        for c2 in col2:
            flat_roll_cor.loc[:, (c + '_' + c2)] = temp.loc[:, c2]

    return flat_roll_cor



import numpy as np

class MeanModel():
    def __init__(self):
        self.model = 'standard_mean'

    def calibrate(self, X, y):  # , method = 'standard'):
        # no hyper parameter to tune
        pass

    def fit(self, X_train, y_train, X_val, y_val):
        # nothing to do
        pass

    def predict(self, X_test):
        y_pred = np.mean(X_test, axis=1).values
        return y_pred

    def get_features_importance(self, features_names):
        run_importances = {}
        # for (name, imp) in zip(features_names, self.model.coef_):
        #     run_importances[name] = imp
        return run_importances


#!/usr/bin/env python3
# coding: utf-8

# Built-in packages

# External packages
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns

# Internal packages
from napoleontoolbox.utility.money_management import iso_vol
from napoleontoolbox.utility.metrics import drawdown, roll_sharpe
from napoleontoolbox.backtest.print_stats import set_text_stats

# Set plot style
plt.style.use('seaborn')

__all__ = ['display_perf']

# =========================================================================== #
#                              Printer Tools                                  #
# =========================================================================== #


def display_perf(
    y_idx, y_est, period=252, title='', params_iv={},
    plot_drawdown=True, plot_roll_sharpe=True, x_axis=None,
    underlying='Underlying', win=252, fees=0,
):
    """ Print dynamic plot of performance indicators (perf, rolling sharpe
    and draw down) of a strategy (raw and iso-volatility) versus its
    underlying.

    Parameters
    ----------
    y_idx : np.ndarray[np.float64, ndim=1]
        Time series of log-returns of the underlying.
    y_est : np.ndarray[np.float64, ndim=1]
        Time series of the signal's strategy.
    period : int, optional
        Number of period per year. Default is 252.
    title : str or list of str, optional
        Title of performance strategy, default is empty.
    plot_drawdown : bool, optional
        If true plot drawdowns, default is True.
    plot_roll_sharpe : bool, optional
        If true plot rolling sharpe ratios, default is True.
    x_axis : list or np.asarray, optional
        x-axis to plot (e.g. list of dates).
    underlying : str, optional
        Name of the underlying, default is 'Underlying'.
    win : int, optional
        Size of the window of rolling sharpe, default is 252.
    fees : float, optional
        Fees to apply at the strategy performance.

    Returns
    -------
    perf_idx : np.ndarray[np.float64, ndim=1]
        Time series of underlying performance.
    perf_est : np.ndarray[np.float64, ndim=1]
        Time series of raw strategy performance.
    perf_ivo : np.ndarray[np.float64, ndim=1]
        Time series of iso-vol strategy performance.

    See Also
    --------
    PlotBackTest, set_text_stats

    """
    if x_axis is None:
        x_axis = range(y_idx.size)

    # Compute returns
    vect_fee = np.zeros(y_est.shape)
    vect_fee[1:] += np.abs(y_est[1:] - y_est[:-1]) * fees
    vect_ret = y_idx * y_est

    # Compute perf.
    perf_idx = np.exp(np.cumsum(y_idx))
    perf_est = np.exp(np.cumsum(vect_ret - vect_fee))
    iv = iso_vol(np.exp(np.cumsum(y_idx)), **params_iv)
    perf_ivo = np.exp(np.cumsum((vect_ret - vect_fee) * iv))

    # Print stats. table
    txt = set_text_stats(
        y_idx, period=period,
        Strategy=y_est,
        Strat_IsoVol=y_est * iv,
        underlying=underlying,
        fees=fees
    )
    print(txt)

    # Plot results
    n = 1 + plot_roll_sharpe + plot_drawdown
    f, ax = plt.subplots(n, 1, figsize=(9, 6), sharex=True)
    if n == 1:
        ax_perf, ax_dd, ax_dd = ax, None, None
        ax_perf.set_xlabel('Date')
    elif n == 2:
        ax_perf = ax[0]
        (ax_dd, ax_roll) = (ax[1], None) if plot_drawdown else (None, ax[1])
        ax[-1].set_xlabel('Date')
    else:
        ax_perf, ax_dd, ax_roll = ax[0], ax[1], ax[2]
    # Plot performances
    ax_perf.plot(
        x_axis,
        100 * perf_est,
        color=sns.xkcd_rgb["pale red"],
        LineWidth=2.
    )
    ax_perf.plot(
        x_axis,
        100 * perf_ivo,
        color=sns.xkcd_rgb["medium green"],
        LineWidth=1.8
    )
    ax_perf.plot(
        x_axis,
        100 * perf_idx,
        color=sns.xkcd_rgb["denim blue"],
        LineWidth=1.5
    )

    # Set notify motion function
    def motion(event):
        N = len(ax_perf.lines[0].get_ydata())
        w, h = f.get_size_inches() * f.dpi - 200
        x = max(event.x - 100, 0)
        j = int(x / w * N)
        ax_perf.legend([
            'Strategy: {:.0f} %'.format(ax_perf.lines[0].get_ydata()[j] - 100),
            'Strat Iso-Vol: {:.0f} %'.format(ax_perf.lines[1].get_ydata()[j] - 100),
            '{}: {:.0f} %'.format(underlying, ax_perf.lines[2].get_ydata()[j] - 100),
        ], loc='upper left', frameon=True, fontsize=10)
        if plot_drawdown is not None:
            ax_dd.legend([
                'Strategy: {:.2f} %'.format(ax_dd.lines[0].get_ydata()[j]),
                'Strat Iso-Vol: {:.2f} %'.format(ax_dd.lines[1].get_ydata()[j]),
                '{}: {:.2f} %'.format(underlying, ax_dd.lines[2].get_ydata()[j]),
            ], loc='upper left', frameon=True, fontsize=10)
        if plot_roll_sharpe is not None:
            ax_roll.legend([
                'Strategy: {:.2f}'.format(ax_roll.lines[0].get_ydata()[j]),
                'Strat Iso-Vol: {:.2f}'.format(ax_roll.lines[1].get_ydata()[j]),
                '{}: {:.2f}'.format(underlying, ax_roll.lines[2].get_ydata()[j]),
            ], loc='upper left', frameon=True, fontsize=10)

    ax_perf.legend(
        ['Strategy', 'Strat Iso-Vol', underlying],
        loc='upper left', frameon=True, fontsize=10
    )
    ax_perf.set_ylabel('Perf.')
    ax_perf.set_yscale('log')
    ax_perf.set_title(title)
    ax_perf.tick_params(axis='x', rotation=30, labelsize=10)
    # Plot DrawDowns
    if plot_drawdown is not None:
        ax_dd.plot(
            x_axis,
            100 * drawdown(perf_est),
            color=sns.xkcd_rgb["pale red"],
            LineWidth=1.4
        )
        ax_dd.plot(
            x_axis,
            100 * drawdown(perf_ivo),
            color=sns.xkcd_rgb["medium green"],
            LineWidth=1.2
        )
        ax_dd.plot(
            x_axis,
            100 * drawdown(perf_idx),
            color=sns.xkcd_rgb["denim blue"],
            LineWidth=1.
        )
        ax_dd.set_ylabel('% DrawDown')
        ax_dd.set_title('DrawDown in percentage')
        ax_dd.tick_params(axis='x', rotation=30, labelsize=10)
    # Plot rolling Sharpe ratio
    if plot_roll_sharpe is not None:
        ax_roll.plot(
            x_axis,
            roll_sharpe(perf_est, period=period, win=win),
            color=sns.xkcd_rgb["pale red"],
            LineWidth=1.4
        )
        ax_roll.plot(
            x_axis,
            roll_sharpe(perf_ivo, period=period, win=win),
            color=sns.xkcd_rgb["medium green"],
            LineWidth=1.2
        )
        ax_roll.plot(
            x_axis,
            roll_sharpe(perf_idx, period=period, win=win),
            color=sns.xkcd_rgb["denim blue"],
            LineWidth=1.
        )
        ax_roll.set_ylabel('Sharpe ratio')
        ax_roll.set_xlabel('Date')
        ax_roll.set_title('Rolling Sharpe ratio')
        ax_roll.tick_params(axis='x', rotation=30, labelsize=10)

    f.canvas.mpl_connect('motion_notify_event', motion)
    plt.show()

    return perf_idx, perf_est, perf_ivo



#!/usr/bin/env python3
# coding: utf-8

""" Object to roll algorithms. """

# Built-in packages

# Third party packages

# Local packages

__all__ = ['_RollingMechanism']

class _EnsemblingRollingMechanism:
    """ Rolling mechanism. """

    def __init__(self, index, n, s):
        """ Initialize object. """
        self.idx = index
        self.n = n
        self.s = s

    def __call__(self, t=None, T=None, display=True):
        self.display = display
        # Set initial index
        if t is None:
            self.t = self.n

        else:
            self.t = t

        # Set max index
        if T is None:
            self.T = self.idx.size

        else:
            self.T = T

        return self

    def __iter__(self):
        # Iterative method
        return self

    def __next__(self):
        # Next method
        t = self.t

        # Display %
        if self.display:
            self._display()

        if self.t >= self.T - 1:

            raise StopIteration

        # Update rolling
        self.t += self.s

        # Set indexes
        self.d_n = self.idx[max(t - self.n, 0)]
        self.d_1 = self.idx[t - self.s]
        self.d = self.idx[t]
        self.d_s = self.idx[min(t + self.s, self.T - 1)]

        return slice(self.d_n, self.d_1), slice(self.d, self.d_s), None

    def _display(self):
        # Display %
        pct = (self.t - self.n - self.s) / (self.T - self.n - self.T % self.s)

        print('{:.2%}'.format(pct), end='\r')

class _RollingMechanism:
    """ Rolling mechanism. """

    def __init__(self, index, n, s):
        """ Initialize object. """
        self.idx = index
        self.n = n
        self.s = s

    def __call__(self, t=None, T=None, display=True):
        self.display = display
        # Set initial index
        if t is None:
            self.t = self.n

        else:
            self.t = t

        # Set max index
        if T is None:
            self.T = self.idx.size

        else:
            self.T = T

        return self

    def __iter__(self):
        # Iterative method
        return self

    def __next__(self):
        # Next method
        t = self.t

        # Display %
        if self.display:
            self._display()

        if self.t >= self.T - 1:

            raise StopIteration

        # Update rolling
        self.t += self.s

        # Set indexes
        self.d_n = self.idx[max(t - self.n, 0)]
        self.d_1 = self.idx[t - 1]
        self.d = self.idx[t]
        self.d_s = self.idx[min(t + self.s-1, self.T - 1)]

        return slice(self.d_n, self.d_1), slice(self.d, self.d_s), None

    def _display(self):
        # Display %
        pct = (self.t - self.n - self.s) / (self.T - self.n - self.T % self.s)

        print('{:.2%}'.format(pct), end='\r')

class _EvalRollingMechanism:
    """ Rolling mechanism. """

    def __init__(self, index, n, s, s_eval):
        """ Initialize object. """
        self.idx = index
        self.n = n
        self.s = s
        self.s_eval = s_eval

    def __call__(self, t=None, T=None, display=True):
        self.display = display
        # Set initial index
        if t is None:
            self.t = self.n
        else:
            self.t = t

        # Set max index
        if T is None:
            self.T = self.idx.size
        else:
            self.T = T

        return self

    def __iter__(self):
        # Iterative method
        return self

    def __next__(self):
        # Next method
        t = self.t

        # Display %
        if self.display:
            self._display()

        if self.t >= self.T - 1:

            raise StopIteration

        # Update rolling
        self.t += self.s

        # Set indexes
        self.d_n = self.idx[max(t - self.n, 0)]
        self.d_1 = self.idx[t - 1]
        self.d_eval_minus_one = self.idx[t - self.s_eval-1]
        self.d_eval = self.idx[t - self.s_eval]
        self.d = self.idx[t]
        self.d_s = self.idx[min(t + self.s-1, self.T - 1)]
        return slice(self.d_n, self.d_eval_minus_one ), slice(self.d, self.d_s), slice(self.d_eval, self.d_1)

    def _display(self):
        # Display %
        pct = (self.t - self.n - self.s) / (self.T - self.n - self.T % self.s)

        print('{:.2%}'.format(pct), end='\r')

class _ExpandingRollingMechanism:
    """ Rolling mechanism. """

    def __init__(self, index, s):
        """ Initialize object. """
        self.idx = index
        self.s = s

    def __call__(self, t=None, T=None, display=True):
        self.display = display
        # Set initial index
        if t is not None:
            self.t = t
        else :
            self.t = 3 * self.s
        # Set max index
        if T is None:
            self.T = self.idx.size
        else:
            self.T = T
        return self

    def __iter__(self):
        # Iterative method
        return self

    def __next__(self):
        # Next method
        t = self.t
        # Display %
        if self.display:
            self._display()

        if self.t >= self.T - 1:
            raise StopIteration
        # Update rolling
        self.t += self.s
        # Set indexes
        self.d_n = self.idx[0]
        self.d_1 = self.idx[t - 1]
        self.d = self.idx[t]
        self.d_s = self.idx[min(t + self.s-1, self.T - 1)]

        return slice(self.d_n, self.d_1), slice(self.d, self.d_s), None

    def _display(self):
        # Display %
        pct = (self.t)/(self.T)
        print('{:.2%}'.format(pct), end='\r')


class _ExpandingEvalRollingMechanism:
    """ Rolling mechanism. """

    def __init__(self, index, s, s_eval):
        """ Initialize object. """
        self.idx = index
        self.s = s
        self.s_eval = s_eval

    def __call__(self, t=None, T=None, display=True):
        self.display = display
        # Set initial index
        if t is not None:
            self.t = t
        else :
            self.t = 2 * self.s_eval
        # Set max index
        if T is None:
            self.T = self.idx.size
        else:
            self.T = T
        return self

    def __iter__(self):
        # Iterative method
        return self

    def __next__(self):
        # Next method
        t = self.t
        # Display %
        if self.display:
            self._display()

        if self.t >= self.T - 1:
            raise StopIteration
        # Update rolling
        self.t += self.s
        # Set indexes
        self.d_n = self.idx[0]
        self.d_eval = self.idx[t - self.s_eval]
        self.d_eval_minus_one = self.idx[t - self.s_eval-1]

        self.d_1 = self.idx[t - 1]
        self.d = self.idx[t]
        self.d_s = self.idx[min(t + self.s-1, self.T - 1)]

        return slice(self.d_n, self.d_eval_minus_one), slice(self.d, self.d_s), slice(self.d_eval,self.d_1)

    def _display(self):
        # Display %
        pct = (self.t)/(self.T)
        print('{:.2%}'.format(pct), end='\r')
